import 'dart:async';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:graphql_flutter_bloc/src/helper.dart';
import 'package:meta/meta.dart';
import 'package:graphql/internal.dart';
import 'package:graphql_flutter/graphql_flutter.dart';

import 'event.dart';
import 'state.dart';

abstract class GraphQLBloc<T> extends Bloc<GraphQLEvent<T>, GraphQLState<T>> {
  GraphQLClient client;
  ObservableQuery result;
  WatchQueryOptions options;

  GraphQLBloc({@required this.client, @required this.options})
      : super(GraphQLState<T>.initial()) {
    result = client.watchQuery(options);

    result.stream.listen((QueryResult result) {
      if (state is GraphQLStateRefetch &&
          result.source == QueryResultSource.Cache) {
        return;
      }

      if (result.loading && result.data == null) {
        add(GraphQLEvent.loading(result: result));
      }

      if (!result.loading && result.data != null) {
        add(
          GraphQLEvent<T>.loaded(
            data: parseData(result.data as Map<String, dynamic>),
            result: result,
          ),
        );
      }

      if (result.hasException) {
        add(GraphQLEvent<T>.error(error: result.exception, result: result));
      }
    });
  }

  void dispose() {
    result.close();
  }

  void run() {
    add(GraphQLEvent<T>.run());
  }

  void refetch() {
    add(GraphQLEvent<T>.refetch());
  }

  bool shouldFetchMore(int i, int threshold) => false;

  bool get isFetchingMore => state is GraphQLStateFetchMore;

  bool get isLoading => state is GraphQLStateLoading;

  bool get isLoaded => state is GraphQLStateLoaded;

  bool get isRefetching => state is GraphQLStateRefetch;

  T parseData(Map<String, dynamic> data);

  bool get hasData => (state is GraphQLStateLoaded<T> ||
      state is GraphQLStateFetchMore<T> ||
      state is GraphQLStateRefetch<T>);

  bool get hasError => state is GraphQLStateError<T>;

  String get getError => hasError
      ? parseOperationException((state as GraphQLStateError<T>).error)
      : null;

  Future<void> _runQuery() async {
    result.fetchResults();
  }

  void _fetchMore(FetchMoreOptions options) {
    result.fetchMore(options);
  }

  void _refetch() => result.refetch();

  @override
  Stream<GraphQLState<T>> mapEventToState(GraphQLEvent<T> event) async* {
    if (event is GraphQLEventRun<T>) {
      _runQuery();
    }

    if (event is GraphQLEventLoading<T>) {
      yield GraphQLState.loading(result: event.result);
    }

    if (event is GraphQLEventLoaded<T>) {
      yield GraphQLState<T>.loaded(data: event.data, result: event.result);
    }

    if (event is GraphQLEventError<T>) {
      yield GraphQLState<T>.error(error: event.error, result: event.result);
    }

    if (event is GraphQLEventRefetch<T> &&
        (state is GraphQLStateLoaded<T> || state is GraphQLStateError<T>)) {
      yield GraphQLState<T>.refetch(
          data: state.maybeWhen(loaded: (data, _) => data, orElse: () => null),
          result: null);
      _refetch();
    }

    if (event is GraphQLEventFetchMore<T> && state is GraphQLStateLoaded<T>) {
      yield GraphQLState<T>.fetchMore(
          data: state.maybeWhen(loaded: (data, _) => data, orElse: () => null),
          result: null);
      _fetchMore(event.options);
    }
  }
}
